/*
 *	dmabound.c
 */

#include <stdlib.h>
#include <dos.h>

#include "global.h"

typedef struct block_header_struct
{
	char		*ptr;
	unsigned long	sig;
} Block_Header;

#define	BLOCK_HEADER_SIG	0xfacecabeL

#define LINEAR_ADDRESS(p)	((((unsigned long) FP_SEG(p)) << 4) + (unsigned long) FP_OFF(p))
#define CROSSES_DMA_PAGE(p,l)	(((LINEAR_ADDRESS(p) + (l) - 1) & 0xffff) < (LINEAR_ADDRESS(p) & 0xffff))

/*
 *  Returns ptr to a buffer that will not cross a 64K DMA Page boundary
 *  Block must be free'd by free_dma_page_aligned_block
 *
 */
void *
malloc_dma_page_aligned_block (size_t size)
{
	char	*block, *p;
	Block_Header	*bhp;

	block = malloc (size + sizeof (Block_Header));
	if (block == NULL)
		return (NULL);

	p = block + sizeof (Block_Header);	/*  our housekeeping overhead  */

	/*
 	 * If it crosses a 64K boundary then (end & 0xffff) < (start & 0xffff)
	 */
	if (CROSSES_DMA_PAGE (p, size))
	{
		free (block);

		if (size > ((65535U - sizeof(Block_Header)) / 2))
			return (NULL);

		block = malloc ((2*size) + sizeof (Block_Header));
		if (block == NULL)
			return (NULL);

		p = block + sizeof (Block_Header);

		if (CROSSES_DMA_PAGE (p, size))
			p += size;	/*  Can't fail for size < 64K  */
	}

	bhp = (Block_Header *) p;
	bhp[-1].ptr = block;
	bhp[-1].sig = BLOCK_HEADER_SIG;

	return (p);
}

void
free_dma_page_aligned_block (void *block)
{
	Block_Header	*bhp;

	if (block == NULL)
		return;

	bhp = (Block_Header *) block;
	if (bhp[-1].sig != BLOCK_HEADER_SIG)
		return;	/*  Bad mem block  */
	free (bhp[-1].ptr);
}

