/*
	DSKINIT.EXE --- Dragon disk formatting utility
	S. Orchard June '99
*/


// V E R S I O N   N U M B E R
#define VER_MAJOR 1
#define VER_MINOR 0


#include <stdio.h>
//#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <ctype.h>
#include <dos.h>

#include <twin.h>

#include "global.h"
#include "hookdbt.h"
#include "dmabound.h"
#include "diskops.h"

// global variables

Byte *buf = NULL;		// disk sector buffer
Byte bam[360];			// bad sector map
int bad_sector_count = 0;
int reformat_count = 0;

// function prototypes

void show_help(void);
int get_drive_num(char *arg);
int format_setup(int drive, int &totaltracks);
int format_disk(int drive, int tracks, int sides);
int reformat_track(int drive, int track, int side);
int verify_disk(int drive, int tracks, int sides);
void flag_sector(int lsn);
int write_dragon_dir(int drive, int tracks, int sides);
void errmsg(const char *msg);
int ctrl_break(void);

// M A I N   F U N C T I O N

int main(int argc, char *argv[])
{
	int i, drive_num = -1, result, errorlevel = 1;

	int totaltracks;		// number of tracks supported on drive

	int tracks = 0,
		sides = 0,
		ex_tracks = 0,
		ex_sides = 0;

	Bool do_format = TRUE,
		 unconditional = FALSE,
		 quick_format = FALSE;

	printf("\nDSKINIT.EXE v%d.%02d\tDragonDos disk format utility\n"
		   "\t\t\tby Stewart Orchard "__DATE__"\n\n", VER_MAJOR, VER_MINOR);

	// Process command line args

	for(i=1; i<argc; i++)
	{
		// Is it a switch?
		if(*argv[i] == '/' || *argv[i] == '-')
		{
			switch(toupper(argv[i][1]))
			{
				case '1' :  sides = 1;
							break;

				case '4' :  tracks = 40;
							break;

				case 'U' :  unconditional = TRUE;
							break;

				case 'Q' :  quick_format = TRUE;
							break;

				case '?' :
				case 'H' :	show_help();
							return 0;

				default  :  printf("Invalid switch '%s'\n", argv[i]);
							return 1;
			}

			/* switch must be single character */
			if(argv[i][2])
			{
				printf("Invalid switch '%s'\n", argv[i]);
				return 1;
			}
		}
		else
		{
			// Must be a drive letter then
			if(drive_num != -1)
			{
				show_help();
				return 1;
			}

			drive_num = get_drive_num(argv[i]);
		}
	}

	if(drive_num == -1)
	{
		show_help();
		return 1;
	}

	// hook in new break handler
	ctrlbrk(ctrl_break);

	buf = (Byte*)malloc_dma_page_aligned_block(256);
	if (buf == NULL)
	{
		printf("\nInternal error: unable to allocate disk buffer\n");
		return 1;
	}

	Window status(0,0,48,12,(GREEN<<4)+LIGHTGREEN,(GREEN<<4)+BLACK);
	_setcursortype(_NOCURSOR);

	if(format_setup(drive_num, totaltracks))
		goto abnormal_exit;

	if(!tracks)
		tracks = totaltracks;

	if(!sides)
		sides = 2;

	memset(bam, 0xff, 360);

	if(!unconditional)
	{
		cprintf("Checking existing format\n\r");
		memset(buf, 0, 256);

		if(!read_sector(drive_num, 20, 0, 1, buf))
		{
			if((buf[252] == (Byte)~buf[254]) && (buf[253] == (Byte)~buf[255]))
			{
				ex_tracks = buf[252];
				ex_sides = ((buf[253] == 18) ? 1 : 2);

				if( ((ex_tracks == 40) || (ex_tracks == 80)) &&
					((ex_sides == 1) || (ex_sides == 2)) )
				{
					if(tracks > ex_tracks)
						tracks = ex_tracks;

					if(sides > ex_sides)
						sides = ex_sides;

					do_format = FALSE;
				}
			}
		}
	}

	if(do_format)
	{
		result = format_track(drive_num, 20, 0);

		if(result)
		{
			errmsg(bios_error_description(result));
			goto abnormal_exit;
		}

		result = verify_sectors(drive_num, 20, 0, 1, 18);

		switch(result)
		{
			case 0x00 : break;

			case 0x02 :
			case 0x04 :
			case 0x10 :	errmsg("Invalid media or track 20 bad - disk cannot be formatted.");
						goto abnormal_exit;

			default   : errmsg(bios_error_description(result));
						goto abnormal_exit;
		}

		cprintf("Formatting\n\r");
		result = format_disk(drive_num, tracks, sides);

		switch(result)
		{
			case 0	:	break;

			case -1 :	cprintf("Format skipped by user\n\r");
						break;

			default :	goto abnormal_exit;
		}
	}
	else    // check for WP
	{
		if(write_sector(drive_num, 0, 0, 1, buf) == 0x03)
		{
			errmsg(bios_error_description(0x03));
			goto abnormal_exit;
		}
	}

	if(!quick_format)
	{
		cprintf("Verifying\n\r");
		result = verify_disk(drive_num, tracks, sides);

		switch(result)
		{
			case 0	:	break;

			case -1 :	cprintf("Verify skipped by user\n\r");
						break;

			default :	goto abnormal_exit;
		}
	}

	cprintf("Writing directory (%d tracks, %d side%s)\n\r", tracks, sides,
			(sides == 1) ? "" : "s");

	if(write_dragon_dir(drive_num, tracks, sides))
		goto abnormal_exit;

	cprintf("Format complete.\n\n\r");

	if(!quick_format)
	{
		cprintf("  %6d  track%s re-formatted during verify\n\r", reformat_count, (reformat_count == 1) ? "" : "s");
		cprintf("  %6ld  bytes in bad sectors\n\r", (long)bad_sector_count << 8);
	}

	cprintf("  %6ld  bytes free", (long)(tracks * sides * 18 - 36 - bad_sector_count) << 8);

	cprintf("\n\n\rPress any key");
	getch();

	errorlevel = 0;

abnormal_exit:

	reset_drive(drive_num);
	free_dma_page_aligned_block(buf);
	_setcursortype(_NORMALCURSOR);

	return errorlevel;
}


// F U N C T I O N S

void show_help(void)
{
	printf("Usage:\n\n"
		   "DSKINIT drive[:] [/1] [/4] [/U] [/Q]\n\n"
		   "/1\tFormat 1 side only.\n"
		   "/4\tFormat 40 tracks only.\n"
		   "/U\tUnconditional format.\n"
		   "/Q\tQuick format (no verification).\n"
		   "/? /H\tShow this help.\n");
}


// converts drive specifier string into drive number
int get_drive_num(char *arg)
{
	if (arg[1] == '\0'				//  Accept 'A', 'A:' or 'A:\'
		|| (arg[1] == ':'
			&& (arg[2] == '\0'
				|| (arg[2] == '\\'
					&& arg[3] == '\0'))))

		switch (*arg)
		{
			case 'a':
			case 'A': return 0;

			case 'b':
			case 'B': return 1;
		}

	return -1;
}


// sets up for format
int format_setup(int drive, int &totaltracks)
{
	int drive_type, result;

	char parms[4][2] = { 1, 40,         // 5.25" in 360K drive
						 2, 40,         // 5.25" in 1.2M   "
						 4, 80,         // 3.5" in 720K    "
						 4, 80 };       // 3.5" in 1.44M   "

	drive_type = bios_get_drive_type(drive) - 1;

	if((drive_type < 0) || (drive_type > 4))
	{
		errmsg("Unsupported drive or drive not installed");
		return 1;
	}

	totaltracks = parms[drive_type][1];

	result = bios_set_dasd_type(drive, parms[drive_type][0]);

	if(result && (result !=6))
	{
		errmsg(bios_error_description(result));
		return 1;
	}

	result = bios_set_media_type(drive, totaltracks-1, 9);

	if(result)
	{
		errmsg(bios_error_description(result));
		return 1;
	}

	return 0;
}

// performs low-level format
int format_disk(int drive, int tracks, int sides)
{
	int track, side,
		count = 0, result, half = 1,
		total = tracks * sides;

	Window win(0,0,31,5,(BLUE<<4)+LIGHTBLUE,(BLUE<<4)+LIGHTGRAY);

	cprintf(" Formatting %d tracks, %d side%s\n", tracks, sides,
			(sides == 1) ? "" : "s");

	gotoxy(3,5);
	cprintf("[press ESC to skip format]");

	textattr((BLACK<<4)+RED);
	gotoxy(6,3);
	cprintf("                    ");
	gotoxy(6,3);

	for(track = 0; track < tracks; track++)
	{
		for(side = 0; side < sides; side++, count += 40)
		{
			if(kbhit() && (getch() == 27))
			{
				return -1;
			}

			if(count >= total)
			{
				count -= total;
				if(half)
				{
					putch('');
					half = 0;
					gotoxy(wherex()-1, wherey());
				}
				else
				{
					putch('');
					half = 1;
				}
			}

			result = format_track(drive, track, side);
			if(result)
			{
				errmsg(bios_error_description(result));
				return 1;
			}
		}
	}
	return 0;
}


// verifies low-level format (tries to re-format bad tracks)
int verify_disk(int drive, int tracks, int sides)
{
	int track, side, sector, lsn,
		count = 0, result, half = 1,
		total = tracks * sides;

	Window win(0,0,30,5,(BLUE<<4)+LIGHTBLUE,(BLUE<<4)+LIGHTGRAY);

	cprintf(" Verifying %d tracks, %d side%s", tracks, sides,
			(sides == 1) ? "" : "s");

	gotoxy(3,5);
	cprintf("[press ESC to skip verify]");

	textattr((BLACK<<4)+GREEN);
	gotoxy(6,3);
	cprintf("                    ");
	gotoxy(6,3);

	for(track = 0; track < tracks; track++)
	{
		for(side = 0; side < sides; side++, count += 40)
		{
			if(kbhit() && (getch() == 27))
			{
				return -1;
			}

			if(count >= total)
			{
				count -= total;
				if(half)
				{
					putch('');
					half = 0;
					gotoxy(wherex()-1, wherey());
				}
				else
				{
					putch('');
					half = 1;
				}
			}

			if(verify_sectors(drive, track, side, 1, 18))
			{
				result = reformat_track(drive, track, side);
				if(result)
				{
					errmsg(bios_error_description(result));
					return 1;
				}

				reformat_count++;

				for(sector = 1; sector <19; sector++)
				{
					if(verify_sectors(drive, track, side, sector, 1))
					{
						lsn = (track * sides + side)*18 + sector - 1;
						flag_sector(lsn);
						bad_sector_count++;
					}
				}
			}
		}
	}
	return 0;
}


int reformat_track(int drive, int track, int side)
{
	char msgbuf[80];

	Window win(0,0,33,1,(LIGHTGRAY<<4)+RED,(LIGHTGRAY<<4)+BLACK);

	sprintf(msgbuf, " re-formatting track %d / side %d", track, side);
	cprintf(msgbuf);

	return format_track(drive, track, side);
}

// flags sector as used in block availability map
void flag_sector(int lsn)
{
	bam[lsn >> 3] &= ~(1 << (lsn & 7));
}


// writes empty Dragon Dos directory
int write_dragon_dir(int drive, int tracks, int sides)
{
	Bool second_bitmap = FALSE;
	int bitmap_size = (tracks * sides * 18) >> 3;

	int	track, dir_lsn, i, result;

	char errmsgbuf[80];

	// 80 track double-sided disks have extended bitmap
	if(bitmap_size > 180)
	{
		bitmap_size = 180;
		second_bitmap = TRUE;
	}

	// allocate 18 sectors for directory on track 20
	dir_lsn = 360 * sides;

	for(i = 0; i<18; i++)
		flag_sector(dir_lsn++);

	// allocate 18 sectors for backup directory on track 16
	dir_lsn = 288 * sides;

	for(i = 0; i<18; i++)
		flag_sector(dir_lsn++);


	// do for tracks 16 & 20
	for(track = 16; track <= 20; track += 4)
	{
		// check integrity of directory track
		if(verify_sectors(drive, track, 0, 1, 18))
		{
			result = reformat_track(drive, track, 0);
			if(result)
			{
				errmsg(bios_error_description(result));
				return 1;
			}
		}

		// set up 1st bitmap sector
		memset(buf, 0, 256);
		memcpy(buf, bam, bitmap_size);

		// set up disk size bytes
		buf[252] = tracks;
		buf[253] = sides * 18;
		buf[254] = ~tracks;
		buf[255] = ~(sides * 18);

		// write 1st bitmap sector
		if((result = write_sector(drive, track, 0, 1, buf)) !=0)
		{
			switch(result)
			{
				case 0x02 :
				case 0x04 :	sprintf(errmsgbuf, "Unable to write to track %d - disk unusable.", track);
							errmsg(errmsgbuf);
							return 1;

				default : errmsg(bios_error_description(result));
						  return 1;
			}
		}

		// set up 2nd bitmap sector
		memset(buf, 0, 256);

		if(second_bitmap)
			memcpy(buf, &bam[180], bitmap_size);

		// write 2nd bitmap sector
		if((result = write_sector(drive, track, 0, 2, buf)) !=0)
		{
			switch(result)
			{
				case 0x02 :
				case 0x04 :	sprintf(errmsgbuf, "Unable to write to track %d - disk unusable.", track);
							errmsg(errmsgbuf);
							return 1;

				default : errmsg(bios_error_description(result));
						  return 1;
			}
		}

		// set up empty directory entries
		memset(buf, 0, 256);

		for(i=0; i<10; i++)
			buf[i*25] = 0x89;

		for(i=3; i<19; i++)
			if((result = write_sector(drive, track, 0, i, buf)) !=0)
			{
				switch(result)
				{
					case 0x02 :
					case 0x04 :	sprintf(errmsgbuf, "Unable to write to track %d - disk unusable.", track);
								errmsg(errmsgbuf);
								return 1;

					default : errmsg(bios_error_description(result));
							  return 1;
				}
			}
	}

	return 0;
}


void errmsg(const char *msg)
{
	int wid = strlen(msg)+2;

	if(wid<18)
		wid = 18;

	Window err(0, 0, wid, 3, (RED<<4)+LIGHTGRAY, (RED<<4)+BLACK);

	gotoxy(2,1);
	cprintf(msg);
	gotoxy((wid>>1)-7,3);
	cprintf("Press any key...");
	getch();
}


// CTRL-BREAK handler puts ESC into keyboard queue
int ctrl_break(void)
{
	ungetch(27);
	return 1;
}

char *copyright = "*** Copyright 1999 - 2000 S.Orchard P.Burgin G.Kinns ***"
				  "*** Original name DSKINIT.EXE ***";