/*
 *	hookdbt.c
 *
 *	Functions for hooking and modifying the BIOS Disk Base Table
 */

#include <stdlib.h>		/*  For a definition of NULL ...  */
#include <dos.h>
#include <mem.h>		/*  movedata()  */

#include "global.h"
#include "hookdbt.h"

Bool
hook_disk_base_table (Disk_Base_Table *dbt)
{
	if (dbt == NULL)
		return (FALSE);

	/*  Get the current Interrupt 1Eh vector  */
	msdos_get_interrupt_vector (DISK_BASE_TABLE_INT_VECTOR,
								&(dbt->orig_segment), &(dbt->orig_offset));

	/*  Copy old table into dbt struct  */
	movedata (dbt->orig_segment, dbt->orig_offset,
			  FP_SEG (dbt->table), FP_OFF (dbt->table), DISK_BASE_TABLE_SIZE);

	/*  Set vector to point at new copy of table  */
	msdos_set_interrupt_vector (DISK_BASE_TABLE_INT_VECTOR,
								FP_SEG (dbt->table), FP_OFF (dbt->table));

	return (TRUE);
}

Bool
unhook_disk_base_table (Disk_Base_Table *dbt)
{
	if (dbt == NULL)
		return (FALSE);

	if (dbt->orig_segment == 0x0000 && dbt->orig_offset == 0x0000)
		return (FALSE);

	/*  Restore Interrupt 1Eh vector  */
	msdos_set_interrupt_vector (DISK_BASE_TABLE_INT_VECTOR,
								dbt->orig_segment, dbt->orig_offset);

	dbt->orig_segment = 0x0000;
	dbt->orig_offset  = 0x0000;

	return (TRUE);
}


void
msdos_get_interrupt_vector (Byte int_num, Word *segment, Word *offset)
{
	union	REGS	regs;
	struct	SREGS	seg_regs;

	/*  Int 21h AH=35h returns vector AL in ES:BX  */
	regs.h.ah = 0x35;
	regs.h.al = int_num;

	int86x (0x21, &regs, &regs, &seg_regs);

	*segment = seg_regs.es;
	*offset  = regs.x.bx;
}

void
msdos_set_interrupt_vector (Byte int_num, Word segment, Word offset)
{
	union	REGS	regs;
	struct	SREGS	seg_regs;

	/*  Int 21h AH=25h sets vector AL to DS:DX  */
	regs.h.ah   = 0x25;
	regs.h.al   = int_num;
	seg_regs.ds = segment;
	regs.x.dx   = offset;

	int86x (0x21, &regs, &regs, &seg_regs);
}
